#include <windows.h>
#include <math.h> //Floating point operations

//Default Parameters
WORD Width=600, Height=400; //Window Dimensions
double XMin=-2, XMax=1, YMin=-1, YMax=1; //The X/Y axis min/max (boundary) points
int MaxIterations=100; //Detail Level
int AnimationFrames=120; //Number of frames to draw a zoom animation in

//Program specific global variables
HDC MainBackBuffer; //This holds the current fractal level

//Function headers
	//Window graphics device functions
	HDC CreateBackBuffer(HWND CopyHwnd); //Creates a backbuffer
	void SizeBackBuffer(HWND CopyHwnd, HDC *BackBuffer); //Sizes a backbuffer's surface to the needed size for the window
	void DrawToWindow(HWND hWnd); //Draw the backbuffer to the main screen
	void AnimateZoom(double NewXMin, double NewXMax, double NewYMin, double NewYMax, HWND hWnd, int Frames); //Animate a zoom in/out. Takes in new boundary points, the window to draw to, and the number of frames to do the animation in
	LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam); //Process windows events
	//Fractal functions
	void CreateFractal(); //Draws the fractal into the backbuffer
	void DrawFractal(DWORD *PixelArray); //Does the actual fractal drawing into a pixel array
	DWORD ProcessColor(int IterationsToEscape, double ValueAtEndReal, double ValueAtEndImaginary, double BailoutRadius, double Power); //Determine the color of a pixel based on its state

int APIENTRY WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR pCmdLine,int nCmdShow)
{
	//Create the window class
	WNDCLASSEX wc = { sizeof(WNDCLASSEX), NULL, WndProc, 0L, 0L, hInst, LoadIcon(hInst, "MAIN_ICON"), LoadCursor(NULL,IDC_ARROW), NULL, NULL, "JuliaFractal", NULL };
	if(RegisterClassEx(&wc)==0)
		return false;

	//Determine the size of the actual window that we want based upon client area dimensions
	RECT WindowRect={0, 0, Width, Height};
	AdjustWindowRect(&WindowRect, WS_OVERLAPPEDWINDOW, false);

	//Create the window and display it
	HWND MainHwnd = CreateWindowEx(NULL, "JuliaFractal", "Julia Fractal", WS_OVERLAPPEDWINDOW, 0, 0, WindowRect.right-WindowRect.left, WindowRect.bottom-WindowRect.top, NULL, NULL, hInst, NULL);
	ShowWindow(MainHwnd, SW_SHOW);
	UpdateWindow(MainHwnd);

	//Draw the main window for the first time
	MainBackBuffer=CreateBackBuffer(MainHwnd);
	CreateFractal();
	DrawToWindow(MainHwnd);

	//Main loop listening for events
	MSG msg;
	ZeroMemory(&msg, sizeof(MSG));
	while(GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	//Destroy windows handles
	DeleteDC(MainBackBuffer);
	UnregisterClass("JuliaFractal", hInst);

	return S_OK;
}

HDC CreateBackBuffer(HWND CopyHwnd) //Creates a backbuffer
{
	HDC CopyDC=GetDC(CopyHwnd);
	HDC NewBackBuffer=CreateCompatibleDC(CopyDC);
	ReleaseDC(CopyHwnd, CopyDC);
	SizeBackBuffer(CopyHwnd, &NewBackBuffer);
	return NewBackBuffer;
}

void SizeBackBuffer(HWND CopyHwnd, HDC *BackBuffer) //Sizes a backbuffer's surface to the needed size for the window
{
	HDC CopyDC=GetDC(CopyHwnd);
	HBITMAP NewBitmap=CreateCompatibleBitmap(CopyDC, Width, Height);
	SelectObject(*BackBuffer, NewBitmap);
	DeleteObject(NewBitmap);
	ReleaseDC(CopyHwnd, CopyDC);
}

void DrawToWindow(HWND hWnd) //Draw the backbuffer to the main screen
{
	//Draw the backbuffer to the main screen
	HDC TheDC=GetDC(hWnd);
	BitBlt(TheDC,0,0,Width,Height,MainBackBuffer,0,0,SRCCOPY);
	ReleaseDC(hWnd, TheDC);
}

void AnimateZoom(double NewXMin, double NewXMax, double NewYMin, double NewYMax, HWND hWnd, int Frames) //Animate a zoom in/out. Takes in new boundary points, the window to draw to, and the number of frames to do the animation in
{
	//Determine zoomed area pixel information
	double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window dimensions (scale) from boundary points
	double NewPixelXMin=(NewXMin-XMin)/ScaleX*Width, NewPixelYMin=(YMax-NewYMax)/ScaleY*Height; //The pixel coordinates for the upper left corner of the zoom section
	double NewPixelWidth=(NewXMax-NewXMin)/ScaleX*Width, NewPixelHeight=(NewYMax-NewYMin)/ScaleY*Height; //The pixel dimensions of the zoom section

	//Copy the current backbuffer to a temporary backbuffer for use in animation, and also create a back buffer for combining during animation
	HDC CombineBackBuffer=CreateBackBuffer(hWnd), TempBackBuffer=CreateBackBuffer(hWnd);
	BitBlt(TempBackBuffer, 0, 0, Width, Height, MainBackBuffer, 0, 0, SRCCOPY);

	//Copy over the new boundary points and draw the new fractal
	XMin=NewXMin;
	XMax=NewXMax;
	YMin=NewYMin;
	YMax=NewYMax;
	CreateFractal();

	//Do the animation
	HDC TheDC=GetDC(hWnd); //The front buffer to draw to
	for(int i=1;i<=Frames;i++)
	{
		//Determine the animation progress percent
		float CurrentProgress=(float)i/Frames;

		//Determine the new pixel position and dimensions for the new fractal
		double NewFractalWidth=(NewPixelWidth+(Width-NewPixelWidth)*CurrentProgress);
		double NewFractalHeight=(NewPixelHeight+(Height-NewPixelHeight)*CurrentProgress);
		double NewFractalX=(NewPixelXMin*(1-CurrentProgress));
		double NewFractalY=(NewPixelYMin*(1-CurrentProgress));

		//Determine how much the new fractal grew and moved
		double GrowX=NewFractalWidth/NewPixelWidth;
		double GrowY=NewFractalHeight/NewPixelHeight;
		double TranslateX=(NewFractalX-NewPixelXMin)/NewPixelWidth;
		double TranslateY=(NewFractalY-NewPixelYMin)/NewPixelHeight;

		//Save the 2 fractal levels to draw so it can be done in different orders
		struct DrawFractalLevel
		{
			int x, y, width, height;
			HDC Buffer;
		};
		DrawFractalLevel F[2]={
			{ //Enlarge the old fractal in proportion to the new fractal
				(int)(Width*TranslateX),
				(int)(Height*TranslateY),
				(int)(Width*GrowX),
				(int)(Height*GrowY),
				TempBackBuffer
			},
			{ //Enlarge the new fractal
				(int)NewFractalX,
				(int)NewFractalY,
				(int)NewFractalWidth,
				(int)NewFractalHeight,
				MainBackBuffer
			}
		};

		//Draw the 2 fractal levels in the proper order
		int SwapI=(GrowX<1)&1; //If zooming out, our grow is <1 . If this is the case,  we will be swapping all "i" instances below to draw in reverse order
		for(int i=0;i<2;i++)
		{
			int n=i^SwapI;
			StretchBlt(CombineBackBuffer, F[n].x, F[n].y, F[n].width, F[n].height, F[n].Buffer, 0, 0, Width, Height, SRCCOPY);
		}

		//Output the combined result
		BitBlt(TheDC, 0, 0, Width, Height, CombineBackBuffer, 0, 0, SRCCOPY);
	}

	//Delete the temporary buffers
	ReleaseDC(hWnd, TheDC);
	DeleteDC(CombineBackBuffer);
	DeleteDC(TempBackBuffer);

	DrawToWindow(hWnd); //Draw the final result
}

void CreateFractal() //Draws the fractal into the MainBackBuffer
{
	//Create the Fractal
	DWORD *FractalArray=new DWORD[Width*Height]; //Create the pixel array
	DrawFractal(FractalArray); //Do the drawing

	//Output to MainBackBuffer
	BITMAPINFO BitmapInfo={sizeof(BITMAPINFOHEADER), Width, Height, 1, 32, BI_RGB, 0, 0, 0, 0, 0};
	SetDIBits(MainBackBuffer, (HBITMAP)GetCurrentObject(MainBackBuffer, OBJ_BITMAP), 0, Height, FractalArray, &BitmapInfo, DIB_RGB_COLORS);

	//Clear out pixel array
	delete[] FractalArray;
}

void DrawFractal(DWORD *PixelArray) //Does the actual fractal drawing into a pixel array
{
	for(int Y=0;Y<Height;Y++)
		for(int X=0;X<Width;X++)
		{
			//Determine the point in the 2D fractal plane we are checking and assign to {x,y} (used for break out testing) and {x0,y0} (which remember the original coordinate)
			double x0=(double)(XMax-XMin)*((double)X/Width)+XMin, y0=(YMax-YMin)*((double)Y/Height)+YMin, x=x0, y=y0;

			//Determine if and when the point breaks out
			for(int iteration=0; x*x+y*y<=(2*2) && iteration<MaxIterations; iteration++)
			{
				double xtemp=x*x-y*y+x0;
				y=2*x*y+y0;
				x=xtemp;
			}

			//Assign the pixel color depending on if and when the point breaks out
			PixelArray[Y*Width+X]=(iteration == MaxIterations ? 0 : ProcessColor(iteration, x, y, 2, 2));
		}
}

DWORD ProcessColor(int IterationsToEscape, double ValueAtEndReal, double ValueAtEndImaginary, double BailoutRadius, double Power) //Determine the color of a pixel based on its state - Parameters = n, Zn{r}, Zn{i} B, P
{
	double x=log(2*log(BailoutRadius))-log(log(sqrt(fabs(ValueAtEndReal*ValueAtEndReal+ValueAtEndImaginary*ValueAtEndImaginary))))/log(Power);
	double State=(double)IterationsToEscape+x;
	return (DWORD)((15-State)/15*0xFF);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) //Process windows events
{
	switch(msg)
	{
		case WM_PAINT:
		{
			//Draw the backbuffer to the main screen
			DrawToWindow(hWnd);
			break;
		}
		case WM_SIZE:
		{
			//Resize the backbuffer
			Width=LOWORD(lParam);
			Height=HIWORD(lParam);
			SizeBackBuffer(hWnd, &MainBackBuffer);

			//Redraw the window
			CreateFractal();
			DrawToWindow(hWnd);
			break;
		}
		case WM_LBUTTONDOWN:
		{
			//Zoom in by 200% around point
			double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window dimensions

			//Determine the clicked point in the window
			double PointX=(double)LOWORD(lParam)/Width*ScaleX+XMin;
			double PointY=YMax-(double)HIWORD(lParam)/Height*ScaleY;

			AnimateZoom(PointX-ScaleX/8, PointX+ScaleX/8, PointY-ScaleY/8, PointY+ScaleY/8, hWnd, AnimationFrames); //Zoom in
			break;
		}
		case WM_RBUTTONDOWN:
		{
			//Zoom out by 200%
			double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window dimensions

			//Determine the center of the window
			double PointX=.5*ScaleX+XMin;
			double PointY=YMax-.5*ScaleY;

			AnimateZoom(PointX-ScaleX*2, PointX+ScaleX*2, PointY-ScaleY*2, PointY+ScaleY*2, hWnd, AnimationFrames); //Zoom out
			break;
		}
		case WM_CHAR:
			if(wParam=='=') //Detail level up
				MaxIterations*=2;
			else if(wParam=='-') //Detail level down
				MaxIterations/=2;
			else //Assumes a direction key was pressed
			{
				//Move in given direction
				double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window dimensions

				//Determine the center of the window
				double PointX=.5*ScaleX+XMin;
				double PointY=YMax-.5*ScaleY;

				if(wParam=='w')
					PointY+=ScaleY/4;
				else if(wParam=='a')
					PointX+=ScaleX/-4;
				else if(wParam=='s')
					PointY+=ScaleY/-4;
				else if(wParam=='d')
					PointX+=ScaleX/4;
				else //If not w/a/s/d, do nothing
					break;

				AnimateZoom(PointX-ScaleX/2, PointX+ScaleX/2, PointY-ScaleY/2, PointY+ScaleY/2, hWnd, AnimationFrames*10); //Slide the window
				break;
			}

			//Redraw the window when changing detail level
			CreateFractal();
			DrawToWindow(hWnd);
			break;
		case WM_DESTROY:
			//Terminate the program
			PostQuitMessage(0);
			return 0;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}